@@ -10,6 +10,7 @@
 //! You may wish to use `termwiz::escape::parser::Parser` in the
 //! [termwiz](https://docs.rs/termwiz/) crate if you don't want to have to research
 //! all those possible escape sequences for yourself.
+#![allow(clippy::upper_case_acronyms)]
 use utf8parse::Parser as Utf8Parser;
 mod enums;
 use crate::enums::*;
@@ -24,7 +25,7 @@ fn lookup(state: State, b: u8) -> (Action, State) {
             .get_unchecked(state as usize)
             .get_unchecked(b as usize)
     };
-    (Action::from_u8(v >> 4), State::from_u8(v & 0xf))
+    (Action::from_u16(v >> 8), State::from_u16(v & 0xff))
 }
 
 #[inline(always)]
@@ -108,6 +109,7 @@ pub trait VTActor {
     /// for more information on device control strings.
     fn dcs_hook(
         &mut self,
+        mode: u8,
         params: &[i64],
         intermediates: &[u8],
         ignored_excess_intermediates: bool,
@@ -148,13 +150,7 @@ pub trait VTActor {
     ///
     /// See [ECMA-48](http://www.ecma-international.org/publications/files/ECMA-ST/ECMA-48,%202nd%20Edition,%20August%201979.pdf)
     /// for more information on control functions.
-    fn csi_dispatch(
-        &mut self,
-        params: &[i64],
-        intermediates: &[u8],
-        ignored_excess_intermediates: bool,
-        byte: u8,
-    );
+    fn csi_dispatch(&mut self, params: &[CsiParam], parameters_truncated: bool, byte: u8);
 
     /// Called when an OSC string is terminated by ST, CAN, SUB or ESC.
     ///
@@ -162,6 +158,10 @@ pub trait VTActor {
     /// that were passed as semicolon separated parameters to the operating
     /// system command.
     fn osc_dispatch(&mut self, params: &[&[u8]]);
+
+    /// Called when an APC string is terminated by ST
+    /// `data` is the data contained within the APC sequence.
+    fn apc_dispatch(&mut self, data: Vec<u8>);
 }
 
 /// `VTAction` is an alternative way to work with the parser; rather
@@ -175,6 +175,7 @@ pub enum VTAction {
         params: Vec<i64>,
         intermediates: Vec<u8>,
         ignored_excess_intermediates: bool,
+        byte: u8,
     },
     DcsPut(u8),
     DcsUnhook,
@@ -185,12 +186,12 @@ pub enum VTAction {
         byte: u8,
     },
     CsiDispatch {
-        params: Vec<i64>,
-        intermediates: Vec<u8>,
-        ignored_excess_intermediates: bool,
+        params: Vec<CsiParam>,
+        parameters_truncated: bool,
         byte: u8,
     },
     OscDispatch(Vec<Vec<u8>>),
+    ApcDispatch(Vec<u8>),
 }
 
 /// This is an implementation of `VTActor` that captures the events
@@ -228,11 +229,13 @@ impl VTActor for CollectingVTActor {
 
     fn dcs_hook(
         &mut self,
+        byte: u8,
         params: &[i64],
         intermediates: &[u8],
         ignored_excess_intermediates: bool,
     ) {
         self.actions.push(VTAction::DcsHook {
+            byte,
             params: params.to_vec(),
             intermediates: intermediates.to_vec(),
             ignored_excess_intermediates,
@@ -262,17 +265,10 @@ impl VTActor for CollectingVTActor {
         });
     }
 
-    fn csi_dispatch(
-        &mut self,
-        params: &[i64],
-        intermediates: &[u8],
-        ignored_excess_intermediates: bool,
-        byte: u8,
-    ) {
+    fn csi_dispatch(&mut self, params: &[CsiParam], parameters_truncated: bool, byte: u8) {
         self.actions.push(VTAction::CsiDispatch {
             params: params.to_vec(),
-            intermediates: intermediates.to_vec(),
-            ignored_excess_intermediates,
+            parameters_truncated,
             byte,
         });
     }
@@ -282,11 +278,15 @@ impl VTActor for CollectingVTActor {
             params.iter().map(|i| i.to_vec()).collect(),
         ));
     }
+
+    fn apc_dispatch(&mut self, data: Vec<u8>) {
+        self.actions.push(VTAction::ApcDispatch(data));
+    }
 }
 
 const MAX_INTERMEDIATES: usize = 2;
 const MAX_OSC: usize = 16;
-const MAX_PARAMS: usize = 16;
+const MAX_PARAMS: usize = 32;
 
 struct OscState {
     buffer: Vec<u8>,
@@ -303,7 +303,7 @@ impl OscState {
                     self.full = true;
                 }
                 num => {
-                    self.param_indices[num - 1] = self.buffer.len();
+                    self.param_indices[num.saturating_sub(1)] = self.buffer.len();
                     self.num_params += 1;
                 }
             }
@@ -329,20 +329,74 @@ pub struct VTParser {
 
     osc: OscState,
 
-    params: [i64; MAX_PARAMS],
+    params: [CsiParam; MAX_PARAMS],
     num_params: usize,
-    current_param: Option<i64>,
+    current_param: Option<CsiParam>,
     params_full: bool,
+    apc_data: Vec<u8>,
 
     utf8_parser: Utf8Parser,
     utf8_return_state: State,
 }
 
+/// Represents a parameter to a CSI-based escaped sequence.
+///
+/// CSI escapes typically have the form: `CSI 3 m`, but can also
+/// bundle multiple values together: `CSI 3 ; 4 m`.  In both
+/// of those examples the parameters are simple integer values
+/// and latter of which would be expressed as a slice containing
+/// `[CsiParam::Integer(3), CsiParam::Integer(4)]`.
+///
+/// There are some escape sequences that use colons to subdivide and
+/// extend the meaning.  For example: `CSI 4:3 m` is a sequence used
+/// to denote a curly underline.  That would be represented as:
+/// `[CsiParam::ColonList(vec![Some(4), Some(3)])]`.
+///
+/// Later: reading ECMA 48, CSI is defined as:
+/// CSI P ... P  I ... I  F
+/// Where P are parameter bytes in the range 0x30-0x3F [0-9:;<=>?]
+/// and I are intermediate bytes in the range 0x20-0x2F
+/// and F is the final byte in the range 0x40-0x7E
+///
+#[derive(Clone, PartialEq, Eq, Debug, Hash)]
+pub enum CsiParam {
+    Integer(i64),
+    P(u8),
+}
+
+impl Default for CsiParam {
+    fn default() -> Self {
+        Self::Integer(0)
+    }
+}
+
+impl CsiParam {
+    pub fn as_integer(&self) -> Option<i64> {
+        match self {
+            Self::Integer(i) => Some(*i),
+            _ => None,
+        }
+    }
+}
+
+impl std::fmt::Display for CsiParam {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            CsiParam::Integer(v) => {
+                write!(f, "{}", v)?;
+            }
+            CsiParam::P(p) => {
+                write!(f, "{}", *p as char)?;
+            }
+        }
+        Ok(())
+    }
+}
+
 impl VTParser {
     #[allow(clippy::new_without_default)]
     pub fn new() -> Self {
         let param_indices = [0usize; MAX_OSC];
-        let params = [0i64; MAX_PARAMS];
 
         Self {
             state: State::Ground,
@@ -359,15 +413,28 @@ impl VTParser {
                 full: false,
             },
 
-            params,
+            params: Default::default(),
             num_params: 0,
             params_full: false,
             current_param: None,
 
             utf8_parser: Utf8Parser::new(),
+            apc_data: vec![],
         }
     }
 
+    fn as_integer_params(&self) -> [i64; MAX_PARAMS] {
+        let mut res = [0i64; MAX_PARAMS];
+        let mut i = 0;
+        for src in &self.params[0..self.num_params] {
+            if let CsiParam::Integer(value) = src {
+                res[i] = *value;
+                i += 1;
+            }
+        }
+        res
+    }
+
     fn finish_param(&mut self) {
         if let Some(val) = self.current_param.take() {
             if self.num_params < MAX_PARAMS {
@@ -377,6 +444,25 @@ impl VTParser {
         }
     }
 
+    /// Promote early intermediates to parameters.
+    /// This is handle sequences such as DECSET that use `?`
+    /// prior to other numeric parameters.
+    /// `?` is technically in the intermediate range and shouldn't
+    /// appear in the parameter position according to ECMA 48
+    fn promote_intermediates_to_params(&mut self) {
+        if self.num_intermediates > 0 {
+            for &p in &self.intermediates[..self.num_intermediates] {
+                if self.num_params >= MAX_PARAMS {
+                    self.ignored_excess_intermediates = true;
+                    break;
+                }
+                self.params[self.num_params] = CsiParam::P(p);
+                self.num_params += 1;
+            }
+            self.num_intermediates = 0;
+        }
+    }
+
     fn action(&mut self, action: Action, param: u8, actor: &mut dyn VTActor) {
         match action {
             Action::None | Action::Ignore => {}
@@ -390,6 +476,7 @@ impl VTParser {
                 self.num_params = 0;
                 self.params_full = false;
                 self.current_param.take();
+                self.apc_data.clear();
             }
             Action::Collect => {
                 if self.num_intermediates < MAX_INTERMEDIATES {
@@ -403,27 +490,39 @@ impl VTParser {
                 if self.params_full {
                     return;
                 }
-                if param == b';' {
-                    if self.num_params + 1 > MAX_OSC {
-                        self.params_full = true;
-                    } else {
-                        self.params[self.num_params] = self.current_param.take().unwrap_or(0);
-                        self.num_params += 1;
-                    }
-                } else {
-                    let current = self.current_param.take().unwrap_or(0);
 
-                    self.current_param.replace(
-                        current
-                            .saturating_mul(10)
-                            .saturating_add((param - b'0') as i64),
-                    );
+                self.promote_intermediates_to_params();
+
+                match param {
+                    b'0'..=b'9' => match self.current_param.take() {
+                        Some(CsiParam::Integer(i)) => {
+                            self.current_param.replace(CsiParam::Integer(
+                                i.saturating_mul(10).saturating_add((param - b'0') as i64),
+                            ));
+                        }
+                        Some(_) => unreachable!(),
+                        None => {
+                            self.current_param
+                                .replace(CsiParam::Integer((param - b'0') as i64));
+                        }
+                    },
+                    p => {
+                        self.finish_param();
+
+                        if self.num_params + 1 > MAX_PARAMS {
+                            self.params_full = true;
+                        } else {
+                            self.params[self.num_params] = CsiParam::P(p);
+                            self.num_params += 1;
+                        }
+                    }
                 }
             }
             Action::Hook => {
                 self.finish_param();
                 actor.dcs_hook(
-                    &self.params[0..self.num_params],
+                    param,
+                    &self.as_integer_params()[0..self.num_params],
                     &self.intermediates[0..self.num_intermediates],
                     self.ignored_excess_intermediates,
                 );
@@ -432,7 +531,7 @@ impl VTParser {
             Action::EscDispatch => {
                 self.finish_param();
                 actor.esc_dispatch(
-                    &self.params[0..self.num_params],
+                    &self.as_integer_params()[0..self.num_params],
                     &self.intermediates[0..self.num_intermediates],
                     self.ignored_excess_intermediates,
                     param,
@@ -440,9 +539,9 @@ impl VTParser {
             }
             Action::CsiDispatch => {
                 self.finish_param();
+                self.promote_intermediates_to_params();
                 actor.csi_dispatch(
                     &self.params[0..self.num_params],
-                    &self.intermediates[0..self.num_intermediates],
                     self.ignored_excess_intermediates,
                     param,
                 );
@@ -475,6 +574,16 @@ impl VTParser {
                 }
             }
 
+            Action::ApcStart => {
+                self.apc_data.clear();
+            }
+            Action::ApcPut => {
+                self.apc_data.push(param);
+            }
+            Action::ApcEnd => {
+                actor.apc_dispatch(std::mem::take(&mut self.apc_data));
+            }
+
             Action::Utf8 => self.next_utf8(actor, param),
         }
     }
@@ -516,7 +625,9 @@ impl VTParser {
                 let byte = ((c as u32) & 0xff) as u8;
 
                 let (action, state) = lookup(self.utf8_return_state, byte);
-                if state != self.utf8_return_state && state != State::Utf8Sequence {
+                if action == Action::Execute
+                    || (state != self.utf8_return_state && state != State::Utf8Sequence)
+                {
                     self.action(lookup_exit(self.utf8_return_state), 0, actor);
                     self.action(action, byte, actor);
                     self.action(lookup_entry(state), 0, actor);
@@ -555,7 +666,7 @@ impl VTParser {
                 self.action(lookup_exit(self.state), 0, actor);
             }
             self.action(action, byte, actor);
-            self.action(lookup_entry(state), 0, actor);
+            self.action(lookup_entry(state), byte, actor);
             self.utf8_return_state = self.state;
             self.state = state;
         } else {
@@ -594,9 +705,8 @@ mod test {
                 VTAction::Print('o'),
                 VTAction::ExecuteC0orC1(0x07,),
                 VTAction::CsiDispatch {
-                    params: vec![32],
-                    intermediates: vec![],
-                    ignored_excess_intermediates: false,
+                    params: vec![CsiParam::Integer(32)],
+                    parameters_truncated: false,
                     byte: b'm',
                 },
                 VTAction::Print('w',),
@@ -604,9 +714,8 @@ mod test {
                 VTAction::Print('o',),
                 VTAction::Print('t',),
                 VTAction::CsiDispatch {
-                    params: vec![0],
-                    intermediates: vec![],
-                    ignored_excess_intermediates: false,
+                    params: vec![CsiParam::Integer(0)],
+                    parameters_truncated: false,
                     byte: b'm',
                 },
                 VTAction::Print('d',),
@@ -647,6 +756,18 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_decset() {
+        assert_eq!(
+            parse_as_vec(b"\x1b[?1l"),
+            vec![VTAction::CsiDispatch {
+                params: vec![CsiParam::P(b'?'), CsiParam::Integer(1)],
+                parameters_truncated: false,
+                byte: b'l',
+            },]
+        );
+    }
+
     #[test]
     fn test_osc_too_many_params() {
         assert_eq!(
@@ -702,15 +823,62 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_fancy_underline() {
+        assert_eq!(
+            parse_as_vec(b"\x1b[4m"),
+            vec![VTAction::CsiDispatch {
+                params: vec![CsiParam::Integer(4)],
+                parameters_truncated: false,
+                byte: b'm'
+            }]
+        );
+
+        assert_eq!(
+            // This is the kitty curly underline sequence.
+            parse_as_vec(b"\x1b[4:3m"),
+            vec![VTAction::CsiDispatch {
+                params: vec![
+                    CsiParam::Integer(4),
+                    CsiParam::P(b':'),
+                    CsiParam::Integer(3)
+                ],
+                parameters_truncated: false,
+                byte: b'm'
+            }]
+        );
+    }
+
+    #[test]
+    fn test_colon_rgb() {
+        assert_eq!(
+            parse_as_vec(b"\x1b[38:2::128:64:192m"),
+            vec![VTAction::CsiDispatch {
+                params: vec![
+                    CsiParam::Integer(38),
+                    CsiParam::P(b':'),
+                    CsiParam::Integer(2),
+                    CsiParam::P(b':'),
+                    CsiParam::P(b':'),
+                    CsiParam::Integer(128),
+                    CsiParam::P(b':'),
+                    CsiParam::Integer(64),
+                    CsiParam::P(b':'),
+                    CsiParam::Integer(192),
+                ],
+                parameters_truncated: false,
+                byte: b'm'
+            }]
+        );
+    }
+
     #[test]
     fn test_csi_omitted_param() {
         assert_eq!(
             parse_as_vec(b"\x1b[;1m"),
             vec![VTAction::CsiDispatch {
-                // The omitted parameter defaults to 0
-                params: vec![0, 1],
-                intermediates: b"".to_vec(),
-                ignored_excess_intermediates: false,
+                params: vec![CsiParam::P(b';'), CsiParam::Integer(1)],
+                parameters_truncated: false,
                 byte: b'm'
             }]
         );
@@ -721,9 +889,41 @@ mod test {
         assert_eq!(
             parse_as_vec(b"\x1b[0;1;2;3;4;5;6;7;8;9;0;1;2;3;4;51;6p"),
             vec![VTAction::CsiDispatch {
-                params: vec![0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 51],
-                intermediates: b"".to_vec(),
-                ignored_excess_intermediates: false,
+                params: vec![
+                    CsiParam::Integer(0),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(1),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(2),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(3),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(4),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(5),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(6),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(7),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(8),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(9),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(0),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(1),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(2),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(3),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(4),
+                    CsiParam::P(b';'),
+                    CsiParam::Integer(51),
+                    CsiParam::P(b';'),
+                ],
+                parameters_truncated: false,
                 byte: b'p'
             }]
         );
@@ -734,28 +934,25 @@ mod test {
         assert_eq!(
             parse_as_vec(b"\x1b[1 p"),
             vec![VTAction::CsiDispatch {
-                params: vec![1],
-                intermediates: b" ".to_vec(),
-                ignored_excess_intermediates: false,
+                params: vec![CsiParam::Integer(1), CsiParam::P(b' ')],
+                parameters_truncated: false,
                 byte: b'p'
             }]
         );
         assert_eq!(
             parse_as_vec(b"\x1b[1 !p"),
             vec![VTAction::CsiDispatch {
-                params: vec![1],
-                intermediates: b" !".to_vec(),
-                ignored_excess_intermediates: false,
+                params: vec![CsiParam::Integer(1), CsiParam::P(b' '), CsiParam::P(b'!')],
+                parameters_truncated: false,
                 byte: b'p'
             }]
         );
         assert_eq!(
             parse_as_vec(b"\x1b[1 !#p"),
             vec![VTAction::CsiDispatch {
-                params: vec![1],
                 // Note that the `#` was discarded
-                intermediates: b" !".to_vec(),
-                ignored_excess_intermediates: true,
+                params: vec![CsiParam::Integer(1), CsiParam::P(b' '), CsiParam::P(b'!')],
+                parameters_truncated: true,
                 byte: b'p'
             }]
         );
@@ -787,4 +984,116 @@ mod test {
             vec![VTAction::Print('\u{af}')]
         );
     }
+
+    #[test]
+    fn utf8_control() {
+        assert_eq!(
+            parse_as_vec("\u{8d}".as_bytes()),
+            vec![VTAction::ExecuteC0orC1(0x8d)]
+        );
+    }
+
+    #[test]
+    fn tmux_control() {
+        assert_eq!(
+            parse_as_vec("\x1bP1000phello\x1b\\".as_bytes()),
+            vec![
+                VTAction::DcsHook {
+                    byte: b'p',
+                    params: vec![1000],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                },
+                VTAction::DcsPut(b'h'),
+                VTAction::DcsPut(b'e'),
+                VTAction::DcsPut(b'l'),
+                VTAction::DcsPut(b'l'),
+                VTAction::DcsPut(b'o'),
+                VTAction::DcsUnhook,
+                VTAction::EscDispatch {
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                    byte: b'\\',
+                }
+            ]
+        );
+    }
+
+    #[test]
+    fn tmux_passthru() {
+        // I'm not convinced that we *should* represent this tmux sequence
+        // in this way, but it is how it currently maps.
+        // It's worth noting that we see this as final byte `t` here, which
+        // collides with decVT105G in https://vt100.net/emu/dcsseq_dec.html
+        assert_eq!(
+            parse_as_vec("\x1bPtmux;data\x1b\\".as_bytes()),
+            vec![
+                VTAction::DcsHook {
+                    byte: b't',
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                },
+                VTAction::DcsPut(b'm'),
+                VTAction::DcsPut(b'u'),
+                VTAction::DcsPut(b'x'),
+                VTAction::DcsPut(b';'),
+                VTAction::DcsPut(b'd'),
+                VTAction::DcsPut(b'a'),
+                VTAction::DcsPut(b't'),
+                VTAction::DcsPut(b'a'),
+                VTAction::DcsUnhook,
+                VTAction::EscDispatch {
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                    byte: b'\\',
+                }
+            ]
+        );
+    }
+
+    #[test]
+    fn kitty_img() {
+        assert_eq!(
+            parse_as_vec("\x1b_Gf=24,s=10,v=20;payload\x1b\\".as_bytes()),
+            vec![
+                VTAction::ApcDispatch(b"Gf=24,s=10,v=20;payload".to_vec()),
+                VTAction::EscDispatch {
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                    byte: b'\\',
+                }
+            ]
+        );
+    }
+
+    #[test]
+    fn sixel() {
+        assert_eq!(
+            parse_as_vec("\x1bPqhello\x1b\\".as_bytes()),
+            vec![
+                VTAction::DcsHook {
+                    byte: b'q',
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                },
+                VTAction::DcsPut(b'h'),
+                VTAction::DcsPut(b'e'),
+                VTAction::DcsPut(b'l'),
+                VTAction::DcsPut(b'l'),
+                VTAction::DcsPut(b'o'),
+                VTAction::DcsUnhook,
+                VTAction::EscDispatch {
+                    params: vec![],
+                    intermediates: vec![],
+                    ignored_excess_intermediates: false,
+                    byte: b'\\',
+                }
+            ]
+        );
+    }
 }
